'''Defines the ZealDB class.'''

# Copyright (c) 2016 Ved Vyas

# This file is part of doxytag2zealdb.

# doxytag2zealdb is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# doxytag2zealdb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with doxytag2zealdb.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function

from builtins import object
import sys
import sqlite3


class ZealDB(object):
    '''
    A ZealDB object can be used to create an SQLite3 database and insert
    entries as expected by Zeal and friends. After constructing a
    ZealDB, one may make multiple calls to insert(...) in between open() and
    close() or while using the ZealDB object as a context manager. Technically,
    the context manager is reusable, and it is similarly OK to have multiple
    pairs open()+close() calls. However, doing this is currently pointless
    because the ``searchIndex'' table is dropped in open().

    Typical usage:

    with ZealDB('/path/to/db.sqlite', verbose=True) as db:
        db.insert('someName', 'someType', 'someFileName')
        ...

    db = ZealDB('/path/to/db.sqlite', verbose=True)

    db.open()
    db.insert(...)
    ...
    db.close()

    # NOTE: see point above about reuse!
    db.open()
    ...
    db.close()

    with db:
        db.insert(...)
        ...
    '''

    def __init__(self, filename, **kwargs):
        '''Initialize a ZealDB.

        A connection is not opened here. To do that, use open() or use this
        ZealDB instance as a context manager.

        Args:
            filename: A string containing full path to database file.
            **verbose: If True, extra information may be printed to stderr
                during operations.
        '''
        self.filename = filename
        self.verbose = kwargs.get('verbose', False)

        self.conn = None
        self.cursor = None

    def open(self):
        '''Open a connection to the database.

        If a connection appears to be open already, transactions are committed
        and it is closed before proceeding. After establishing the connection,
        the searchIndex table is prepared (and dropped if it already exists).
        '''
        if self.conn is not None:
            self.close()

        self.conn = sqlite3.connect(self.filename)
        self.cursor = self.conn.cursor()

        c = self.cursor
        c.execute('SELECT name FROM sqlite_master WHERE type="table"')
        if (u'searchIndex',) in c:
            c.execute('DROP TABLE searchIndex')

            if self.verbose:
                print('Dropped existing table', file=sys.stderr)

        c.executescript(
            '''
            CREATE TABLE searchIndex
            (id INTEGER PRIMARY KEY, name TEXT, type TEXT, path TEXT);

            CREATE UNIQUE INDEX anchor ON searchIndex (name, type, path);
            '''
        )

    def close(self):
        '''Commit transactions and close connection if open.'''
        if self.conn is not None:
            self.conn.commit()
            self.conn.close()

        self.cursor = None
        self.conn = None

    def __enter__(self):
        '''Enter ZealDB context. Opens a connection to the database.'''
        self.open()

        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        '''Exit ZealDB context. Commits transactions and closes the database
        connection.'''
        if exc_type is not None:
            print('Exception occurred, committing changes and closing DB...',
                  file=sys.stderr)

        self.close()

    def insert(self, name, entry_type, filename):
        '''Insert an entry into the Zeal database.

        Args:
            name: A string representing the name of the entry.
            entry_type: A string representing the entry type.
            filename: A string representing the filename of the documentation
                for the entry.

        Raises:
            RuntimeError: a database connection was not established before
                calling insert()
        '''
        if self.cursor is None:
            raise RuntimeError(
                'Open DB connection before attempting to call insert!')

        db_entry = (name, entry_type, filename)

        if self.verbose:
            print('Inserting %s "%s" -> %s' % db_entry, file=sys.stderr)

        self.cursor.execute(
            '''INSERT OR IGNORE INTO searchIndex(name, type, path)
               VALUES (?, ?, ?)''', db_entry)



'''Defines the TagProcessor base class and useful, concrete, child classes.'''

# Copyright (c) 2016 Ved Vyas

# This file is part of doxytag2zealdb.

# doxytag2zealdb is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# doxytag2zealdb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with doxytag2zealdb.  If not, see <http://www.gnu.org/licenses/>.

from builtins import str
from builtins import object


class TagProcessor(object):
    '''
    TagProcessor is a base class for finding specific tags in a BeautifulSoup
    and then extracting information from them to insert into a Zeal
    database. TagProcessors are registered with and then executed by a
    TagfileProcessor.

    The key functionality provided by a TagProcessor is:

    1. A match_criterion method to identify tags of interest
    2. A find generator to yield soup tags that match this criterion
    3. Three methods, get_name, get_entry_type, and get_filename, to extract
    information from the tag that is needed for a Zeal database.

    This base class provides default implementations where possible. Child
    classes should make sure all three categories above are implemented.

    When creating a new TagProcessor, be sure to register it with
    TagfileProcessor.register_tag_processor(...). If it's generally useful, add
    it to the standard registrations in
    TagfileProcessor.init_tag_processors(...).
    '''

    def __init__(self, **kwargs):
        '''Initializer.

        Args:
            **include_parent_scopes: bool. See get_name() for more info.
        '''
        self.include_parent_scopes = kwargs.get('include_parent_scopes', False)

    def match_criterion(self, tag):
        '''Determine if a tag matches a particular criterion.

        Override this to define your own criterion.

        Args:
            tag: A BeautifulSoup Tag to test against the criterion.

        Returns:
            True if the tag satifies the criterion, otherwise False.
        '''
        pass

    def find(self, soup):
        '''Yield tags matching the tag criterion from a soup.

        There is no need to override this if you are satisfied with finding
        tags that match match_criterion.

        Args:
            soup: A BeautifulSoup to search through.

        Yields:
            BeautifulSoup Tags that match the criterion.
        '''
        for tag in soup.recursiveChildGenerator():
            if self.match_criterion(tag):
                yield tag

    def get_name(self, tag):
        '''Extract and return a representative "name" from a tag.

        Override as necessary. get_name's output can be controlled through
        keyword arguments that are provided when initializing a
        TagProcessor. For instance, a member of a class or namespace can have
        its parent scope included in the name by passing
        include_parent_scopes=True to __init__().

        Args:
            tag: A BeautifulSoup Tag that satisfies match_criterion.

        Returns:
            A string that would be appropriate to use as an entry name in a
            Zeal database.
        '''
        name = tag.findChild('name').contents[0]

        if self.include_parent_scopes:
            # Include parent scope in returned name
            parent_tag = tag.findParent()
            if parent_tag.get('kind') in ['class', 'struct', 'namespace']:
                name = parent_tag.findChild('name').contents[0] + '::' + name

        return name

    def get_entry_type(self, tag):
        '''Extract and return a representative "entry type" from a tag.

        Override as necessary.

        Args:
            tag: A BeautifulSoup Tag that satisfies match_criterion.

        Returns:
            A string that would be appropriate to use as an entry type in a
            Zeal database.
        '''
        pass

    def get_filename(self, tag):
        '''Extract and return a documentation filename from a tag.

        Override as necessary, though this default implementation probably
        covers all the cases of interest.

        Args:
            tag: A BeautifulSoup Tag that satisfies match_criterion.

        Returns:
            A string that would be appropriate to use as the documentation
            filename for an entry in a Zeal database.
        '''
        if tag.find('filename', recursive=False) is not None:
            return tag.filename.contents[0]
        elif tag.find('anchorfile', recursive=False) is not None:
            return tag.anchorfile.contents[0] + '#' + tag.anchor.contents[0]


class TagProcessorWithEntryTypeAndFindByNamePlusKind(TagProcessor):
    '''A convenient TagProcessor subclass for cases where get_entry_type should
    return the same thing for all (matching) tags and where matching tags are
    identified by their name and "kind" attribute.
    '''

    def __init__(self, entry_type, tag_name, tag_kind, **kwargs):
        '''Initializer.

        Args:
            entry_type: A string that should be returned by get_entry_type()
                for all (matching) tags.
            tag_name: The unicode string name that matching tags should have.
            tag_kind: The unicode string "kind" attribute that matching tags
                should have.
        '''
        super(TagProcessorWithEntryTypeAndFindByNamePlusKind,
              self).__init__(**kwargs)

        # Save this to return from get_entry_type override
        self.entry_type = entry_type

        # Save these for use in match_criterion override
        self.reference_tag_name = tag_name
        self.reference_tag_kind = tag_kind

    def match_criterion(self, tag):
        '''Override. Determine if a tag has the desired name and kind attribute
        value.

        Args:
            tag: A BeautifulSoup Tag.

        Returns:
            True if tag has the desired name and kind, otherwise False.
        '''
        return tag.name == self.reference_tag_name and \
            tag.attrs.get('kind', '') == self.reference_tag_kind

    def get_entry_type(self, tag):
        '''Override. Return entry type this instance was initialized with.

        Args:
            tag: A BeautifulSoup Tag that satisfies match_criterion.

        Returns:
            string entry type that this instance was initialized with.
        '''
        return self.entry_type


class TagProcessorWithAutoEntryTypeAndFindByNamePlusAutoKind(
        TagProcessorWithEntryTypeAndFindByNamePlusKind):
    '''
    This class attempts to provide further convenience by automatically
    determining the entry type and tag kind attribute from the class name. So
    subclass "fooTagProcessor" will have "Foo" automatically extracted as the
    entry type and "foo" automatically extracted as the tag kind attribute when
    it calls its superclass initializer.
    '''

    def __init__(self, tag_name, **kwargs):
        '''Initializer.

        Args:
            tag_name: unicode string name of tag to match. Usually u'compound'
                or u'member'.
        '''
        # Extract entry type from class name, assuming that it is of the format
        # "$(entry_type)TagProcessor"
        class_name = type(self).__name__
        end_idx = class_name.rfind('TagProcessor')

        tag_kind = str(class_name[:end_idx])
        entry_type = tag_kind.capitalize()

        # Pass it off to the superclass initializer and we're done
        super(TagProcessorWithAutoEntryTypeAndFindByNamePlusAutoKind,
              self).__init__(entry_type, tag_name, tag_kind, **kwargs)


class TagProcessorWithAutoStuffAndCompoundTagName(
        TagProcessorWithAutoEntryTypeAndFindByNamePlusAutoKind):
    '''
    This subclass goes a step further and hard-codes the name of matching tags
    to u"compound".
    '''

    def __init__(self, **kwargs):
        super(TagProcessorWithAutoStuffAndCompoundTagName, self).__init__(
            u'compound', **kwargs)


class TagProcessorWithAutoStuffAndMemberTagName(
        TagProcessorWithAutoEntryTypeAndFindByNamePlusAutoKind):
    '''
    This subclass goes a step further and hard-codes the name of matching tags
    to u"member".
    '''

    def __init__(self, **kwargs):
        super(TagProcessorWithAutoStuffAndMemberTagName, self).__init__(
            u'member', **kwargs)


class classTagProcessor(TagProcessorWithAutoStuffAndCompoundTagName):
    '''Process class tags.'''
    pass


class fileTagProcessor(TagProcessorWithAutoStuffAndCompoundTagName):
    '''Process file tags.'''

    def get_filename(self, tag):
        '''Override. Get correct documentation filename from a file tag.

        Args:
            tag: A BeautifulSoup Tag for a file.

        Returns:
            A string containing the correct documentation filename from the
            tag.
        '''
        return super(fileTagProcessor, self).get_filename(tag) + '.html'


class namespaceTagProcessor(TagProcessorWithAutoStuffAndCompoundTagName):
    '''Process namespace tags.'''
    pass


class structTagProcessor(TagProcessorWithAutoStuffAndCompoundTagName):
    '''Process struct tags.'''
    pass


class unionTagProcessor(TagProcessorWithAutoStuffAndCompoundTagName):
    '''Process union tags.'''
    pass


class functionTagProcessor(TagProcessorWithAutoStuffAndMemberTagName):
    '''Process function tags.'''

    def __init__(self, **kwargs):
        '''Initializer.

        Args:
            **include_function_signatures: bool. See get_name() for more info.
        '''
        super(functionTagProcessor, self).__init__(**kwargs)

        self.include_function_signatures = kwargs.get(
            'include_function_signatures', False)

    def get_name(self, tag):
        '''Override. Extract a representative "name" from a function tag.

        get_name's output can be controlled through keyword arguments that are
        provided when initializing a functionTagProcessor. For instance,
        function arguments and return types can be included by passing
        include_function_signatures=True to __init__().

        Args:
            tag: A BeautifulSoup Tag for a function.

        Returns:
            A string that would be appropriate to use as an entry name for a
            function in a Zeal database.
        '''
        name = super(functionTagProcessor, self).get_name(tag)

        if self.include_function_signatures:
            # Include complete function signature in returned name
            func_args = tag.findChild('arglist')
            if func_args and len(func_args.contents):
                name += func_args.contents[0]

            ret_type = tag.findChild('type')
            if ret_type and len(ret_type.contents):
                name += ' -> ' + ret_type.contents[0]

        return name

    def get_entry_type(self, tag):
        '''Override that returns u'Method' for class/struct methods.

        Override as necessary.

        Args:
            tag: A BeautifulSoup Tag for a function.

        Returns:
            If this is a class/struct method, returns u'Method', otherwise
            returns the value from the inherited implementation of
            get_entry_type (which should be u'Function').
        '''

        if tag.findParent().get('kind') in ['class', 'struct']:
            return u'Method'

        return super(functionTagProcessor, self).get_entry_type(tag)


class defineTagProcessor(TagProcessorWithAutoStuffAndMemberTagName):
    '''Process #define tags.'''
    pass


class enumerationTagProcessor(TagProcessorWithAutoStuffAndMemberTagName):
    '''Process enumeration tags.'''

    def __init__(self, **kwargs):
        '''Initializer. Utilizes inherited machinery, then manually tweaks the
        entry type to be one of the types supported by Dash (see README).'''
        super(enumerationTagProcessor, self).__init__(**kwargs)
        self.entry_type = u'Enum'


class enumvalueTagProcessor(TagProcessorWithAutoStuffAndMemberTagName):
    '''Process enumeration value tags.'''

    def __init__(self, **kwargs):
        '''Initializer. Utilizes inherited machinery, then manually tweaks the
        entry type to be one of the types supported by Dash (see README).'''
        super(enumvalueTagProcessor, self).__init__(**kwargs)
        self.entry_type = u'Value'


class typedefTagProcessor(TagProcessorWithAutoStuffAndMemberTagName):
    '''Process typedef tags.'''

    def __init__(self, **kwargs):
        '''Initializer. Utilizes inherited machinery, then manually tweaks the
        entry type to be one of the types supported by Dash (see README).'''
        super(typedefTagProcessor, self).__init__(**kwargs)
        self.entry_type = u'Type'


class variableTagProcessor(TagProcessorWithAutoStuffAndMemberTagName):
    '''Process variable tags.'''
    pass



'''Defines the DoxygenPropertyList class for operating on Info.plist in docsets
'''

# Copyright (c) 2018 Ved Vyas

# This file is part of doxytag2zealdb.

# doxytag2zealdb is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# doxytag2zealdb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with doxytag2zealdb.  If not, see <http://www.gnu.org/licenses/>.

from bs4 import BeautifulSoup


class DoxygenPropertyList(object):
    '''
    A DoxygenPropertyList object can be used to do basic operations on a
    property list file generated by Doxygen when preparing HTML docsets. Such
    property lists contain a top-level dictionary with various key-value
    pairs. DoxygenPropertyList can then add new key-value pairs, modify
    existing ones, and finally save changes to the original file. This class is
    pretty limited at the moment -- it can't handle general property list
    files and only supports string, integer, and Boolean values. Also, it is
    possible that the formatting and ordering of the original property list
    won't be maintained when changes are saved.

    Typical usage:

    plist = DoxygenPropertyList('/path/to/Info.plist')

    # The following is an example of changing an existing key-value
    plist.set_property('CFBundleName', 'MyCoolProject')
    plist.set_property('SomeIntegerProperty', 2)
    plist.set_property('SomeBooleanProperty', False)

    plist.save()
    '''

    def __init__(self, plist_filename):
        '''Initialize a property list representation from an existing file.

        Args:
            plist_filename: A string containing the full path to a
            Doxygen-generated property list file.

        Raises:
            OSError / FileNotFoundError: Input file cannot be read
            RuntimeError: The property list file is not of the expected format
        '''
        self.filename = plist_filename

        with open(self.filename, 'r') as plist_file:
            self.soup = BeautifulSoup(plist_file, 'lxml-xml')
            self.properties = self.soup.findChild(name='dict')

            if self.properties is None:
                raise RuntimeError('Invalid property list file provided')

    def set_property(self, key, value):
        '''Set a new (or updating existing) key value pair.

        Args:
            key: A string containing the key namespace
            value: A str, int, or bool value

        Raises:
            NotImplementedError: an unsupported value-type was provided
        '''
        value_type = type(value)
        if value_type not in [str, int, bool]:
            raise NotImplementedError(
                'Only string, integer, and boolean properties are implemented')

        key_object = self.properties.findChild(name='key', text=key)

        # Key (and value, if it's a valid property list) don't exist
        if key_object is None:
            key_object = self.soup.new_tag('key')
            key_object.string = key

            self.properties.append(key_object)

            value_object = self.soup.new_tag(
                {str: 'string', int: 'integer', bool: str(value).lower()}[
                    value_type])
            if value_type is not bool:
                value_object.string = str(value)

            self.properties.append(value_object)

            return

        # Key (and value, if it's a valid property list) exist
        # Eh, just remove the key+value tags from the tree and re-add them
        # (with the new value)
        value_object = key_object.find_next_sibling()

        key_object.decompose()
        value_object.decompose()

        self.set_property(key, value)

    def save(self):
        '''Save current property list representation to the original file.'''
        with open(self.filename, 'w') as plist_file:
            plist_file.write(str(self.soup))



#!/usr/bin/python
'''
doxytag2zealdb: creates a SQLite3 database from a Doxygen tag file to enable
searchable Doxygen docsets with categorized entries in tools like helm-dash
[1], Zeal [2], and Dash [3]

Copyright (c) 2016-2019 Ved Vyas

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

Usage:
  doxytag2zealdb [-v] --tag FILENAME --db FILENAME
                 [--include-parent-scopes]
                 [--include-function-signatures]
                 [--dont-update-info-plist]
  doxytag2zealdb (-h | --help | --version)

Options:
  -h --help                      Show this screen
  --version                      Show version
  --tag FILENAME                 Input Doxygen tag file to process
  --db FILENAME                  Output SQlite3 database
  -v --verbose                   Print further information while processing the
                                 tag file
  --include-parent-scopes        Include parent scope in entry names
  --include-function-signatures  Include function arguments and return types in
                                 entry names
  --dont-update-info-plist       Don't automatically update Info.plist for Dash

See the README for further information on how to use doxytag2zealdb while
preparing docsets from Doxygen output.

References:
[1]: https://github.com/areina/helm-dash
[2]: https://github.com/zealdocs/zeal
[3]: https://kapeli.com/dash
'''

from __future__ import print_function

from docopt import docopt
import os

from . import __version__
from .doxytagfile import TagfileProcessor
from .zealdb import ZealDB
from .propertylist import DoxygenPropertyList


def main():
    args = docopt(__doc__, version='doxytag2zealdb v' + __version__)

    verbose = args.get('--verbose', False)
    tag_filename = args['--tag']
    db_filename = args['--db']

    # Prepare keyword arguments that should propagate all the way to
    # TagProcessors
    opts = {}
    for opt in [
            '--include-parent-scopes',
            '--include-function-signatures'
    ]:
        # Make valid kwarg names by stripping leading '-'s or '--'s and
        # replacing other '-'s with underscores
        opt_kwarg_name = opt.strip('-').replace('-', '_')

        # Pass whatever docopt reports, not just if it's non-False
        opts[opt_kwarg_name] = args[opt]

    with ZealDB(db_filename, verbose=verbose) as zdb:
        with open(tag_filename, 'r') as tag:
            tagfile_proc = TagfileProcessor(tag, zdb, verbose=verbose, **opts)
            tagfile_proc.process()

    if not args['--dont-update-info-plist']:
        plist_filename = os.path.join(
            os.path.dirname(os.path.dirname(db_filename)), 'Info.plist')

        plist = DoxygenPropertyList(plist_filename)
        plist.set_property('isDashDocset', True)
        plist.save()



from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

about = {}
with open(path.join(here, 'doxytag2zealdb', '__about__.py')) as fp:
    exec(fp.read(), about)

__version__ = about['__version__']
PROJECT_URL = 'https://gitlab.com/vedvyas/doxytag2zealdb'

setup(
    # TODO: move all of this metadata to doxytag2zealdb/__about__.py?
    name='doxytag2zealdb',
    version=__version__,

    description='''doxytag2zealdb creates a SQLite3 database from a Doxygen tag
    file to enable searchable Doxygen docsets with categorized entries in tools
    like helm-dash, Zeal, and Dash.''',
    long_description=open(path.join(here, 'README.md'),
                          encoding='utf-8').read(),
    long_description_content_type='text/markdown',
    keywords='Zeal Dash Doxygen SQLite3',

    license='GPLv3+',

    author='Ved Vyas',
    author_email='ved@vyas.io',

    url=PROJECT_URL,
    download_url='%s/repository/archive.tar.bz2?ref=v%s' % (PROJECT_URL,
                                                            __version__),

    classifiers=[
        'Development Status :: 4 - Beta',

        'Intended Audience :: Developers',
        'Topic :: Software Development :: Documentation',

        'License :: OSI Approved :: ' +
        'GNU General Public License v3 or later (GPLv3+)',

        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3'
    ],

    install_requires=[
        'docopt >= 0.6.2',
        'lxml >= 3.6.0',
        'beautifulsoup4 >= 4.4.1',
        'future'
    ],
    packages=find_packages(exclude=['docs', 'tests*']),
    package_data={
        'doxytag2zealdb': ['README.md', 'COPYING', 'CONTRIBUTORS']
    },

    entry_points={
        'console_scripts': [
            'doxytag2zealdb = doxytag2zealdb.doxytag2zealdb:main'
        ]
    }
)



'''Defines the TagfileProcessor class.'''

# Copyright (c) 2016 Ved Vyas

# This file is part of doxytag2zealdb.

# doxytag2zealdb is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# doxytag2zealdb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with doxytag2zealdb.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function

from builtins import object
import sys
from bs4 import BeautifulSoup

from .doxytag import *


class TagfileProcessor(object):
    '''
    A TagfileProcessor object can be used to run through a Doxygen-generated
    tag file and insert entries into a Zeal database (via ZealDB). It depends
    on TagProcessors to find tags of interest in the file and extract the
    necessary information from them for insertion into the DB. During
    initialization, a handful of TagProcessors are registered for automatic use
    during a call to process().

    TagfileProcessor manages the BeautifulSoup object used to traverse the tag
    file. It passes this soup and individual tags to TagProcessors to work
    with.

    Typical usage:

    with ZealDB(db_filename, verbose=verbose) as db:
        with open(tag_filename, 'r') as tag:
            tagfile_proc = TagfileProcessor(tag, db, verbose=verbose)
            tagfile_proc.process()
    '''

    def __init__(self, file_obj, zeal_db, **kwargs):
        '''Initialize a tag file processor.

        Args:
            file_obj: A file-like object to read the tags from.
            zeal_db: A ZealDB instance that should be open() by the time
                process is called on this instance.
            **verbose: If True, extra information may be printed to stderr
                during operations.
            **kwargs: **verbose and other keyword arguments _may_ be passed on
                to TagProcessors that are registered in
                init_tag_processors(). Refer to that method and various
                TagProcessor initializers to see which keyword arguments are
                useful.
        '''
        self.soup = BeautifulSoup(file_obj, 'lxml-xml')
        self.zeal_db = zeal_db

        self.opts = kwargs
        self.verbose = kwargs.get('verbose', False)

        self.entry_count = 0

        self.tag_procs = {}
        self.init_tag_processors()

    def init_tag_processors(self):
        '''Register the TagProcessors that are bundled with doxytag2zealdb.'''
        register = self.register_tag_processor

        register('class', classTagProcessor(**self.opts))
        register('file', fileTagProcessor(**self.opts))
        register('namespace', namespaceTagProcessor(**self.opts))
        register('struct', structTagProcessor(**self.opts))
        register('union', unionTagProcessor(**self.opts))
        register('function', functionTagProcessor(**self.opts))
        register('define', defineTagProcessor(**self.opts))
        register('enumeration', enumerationTagProcessor(**self.opts))
        register('enumvalue', enumvalueTagProcessor(**self.opts))
        register('typedef', typedefTagProcessor(**self.opts))
        register('variable', variableTagProcessor(**self.opts))

    def register_tag_processor(self, name, tag_processor):
        '''Register a tag processor.

        Args:
            name: A string to help keep track of the tag processor. Used as a
                key to access the tag processor.
            tag_processor: A TagProcessor instance to associate with ``name''.
        '''
        self.tag_procs[name] = tag_processor

    def unregister_tag_processor(self, name):
        '''Unregister a tag processor.

        Args:
            name: A string key that maps to the TagProcessor to unregister.
        '''
        if name in self.tag_procs:
            del self.tag_procs[name]

    def process(self):
        '''Run all tag processors.'''
        for tag_proc in self.tag_procs:
            before_count = self.entry_count
            self.run_tag_processor(tag_proc)
            after_count = self.entry_count

            if self.verbose:
                print('Inserted %d entries for "%s" tag processor' % (
                    after_count - before_count, tag_proc), file=sys.stderr)

        if self.verbose:
            print('Inserted %d entries overall' % self.entry_count,
                  file=sys.stderr)

    def run_tag_processor(self, tag_proc_name):
        '''Run a tag processor.

        Args:
            tag_proc_name: A string key that maps to the TagProcessor to run.
        '''
        tag_processor = self.tag_procs[tag_proc_name]

        for tag in tag_processor.find(self.soup):
            self.process_tag(tag_proc_name, tag)

    def process_tag(self, tag_proc_name, tag):
        '''Process a tag with a tag processor and insert a DB entry.

        Args:
            tag_proc_name: A string key that maps to the TagProcessor to use.
            tag: A BeautifulSoup Tag to process.
        '''
        tag_processor = self.tag_procs[tag_proc_name]

        db_entry = (tag_processor.get_name(tag),
                    tag_processor.get_entry_type(tag),
                    tag_processor.get_filename(tag))

        self.zeal_db.insert(*db_entry)

        self.entry_count += 1
